; When you ll try to get source for assembling on CP/M, uncomment this line,
; find one more include in the text and replace them with #include
; and run via Linux cpp
; #include "defs.h"
;
; Do not forger to Edit MEMRGN!
;
; Then, delete all lines started with '# '.
; Next, add ^M, ^Z, and change all the ';'s by '*' and you are done ;-)
;
;
; Part I - assembling options
;
; define if BIOS should be compiled as SRecord stand-alone file
AS_SRECORD	equ	0

; create an CPMLDR bios instead
AS_CPMLDR_BIOS	equ	0

; Without this, BIOS will write and read every CP/M sector by
; reading / writing the whole SD sector.
; I do not use the true deblocking algorythm, but this setting will prevent
; reading sectors if they are allready in memory

SD_READ_CACHE	equ	1

; Some debug options 

; Traces all bios calls ;-) (except 1 2 3 4)
TRACE_ALL_BIOS_CALLS	equ	0

;Writes info of sector read and written
SD_SECTOR_DEBUG	equ	0

; Huge write debug
SD_WRITE_DEBUG	equ	0

; Bios messages on what has failed
SD_VERBOSE_ERRMSGS	equ	1

; Issue CMD13 after sector write (possibly wrong and uncomplete)
SD_CMD13_AFTER_WRITE	equ	0


;
; Part II - CP/M settings
;
; Max Disks
; Only one for CPMLDR
	IFNE AS_CPMLDR_BIOS
maxdsk	equ	1
	ENDC
; As much as we can for system itself
	IFEQ AS_CPMLDR_BIOS
maxdsk	equ	16
	ENDC

; Length of DPH (Refer to Sysguide)
dphlen	equ	26

;
; Part III - Hardware defs
;
; this defines specific to 68kavr
uartstat	equ $FA0002
uartdata	equ $FA0000
slowclock	equ $FC0000
fastclock	equ $FE0000
sd_port		equ $F80000
SD_CMD17	equ $51
SD_CMD13	equ $4D
SD_CMD24	equ $58

; END OF SETTINGS

; if we are assembled as SRECORD, all the ORGS are absolute
; (taken from CPM15000.MAP file)
	IFNE AS_SRECORD
	org	$1B000
; For 1.3 
_ccp	equ	$150BC
; For 1.2 _ccp	equ	$150B8
;_init	equ	$1B000
	ENDC

; if not - compile as .o
	IFEQ AS_SRECORD
	.globl	_init
	IFNE	AS_CPMLDR_BIOS	; CPMLDR Does not need _ccp
	.globl	_ccp
	ENDC
;	section	0
	ENDC

;************************************ ENTRY POINT ***********************************

;********* Normal BIOS init
	IFEQ	AS_CPMLDR_BIOS	
; This code is directly from "CP/M 68 System Guide" page 59
	EVEN
; if srecord, _init is defined in CPM15000.SR, so we do not need define it here
	IFNE AS_SRECORD
_init:
	ENDC
	move.l	#biosmsg,a0
	bsr	tx_str
	move.l	#entry,($8c)	
; installing trap #3 handler , so I suppose CP/M talks with bios 
; with this trap
	clr.l	d0	; log on disk A,user 0
	rts
	ENDC
;not an cpmldr bios

;********** CPMLDR BIOS
	IFNE	AS_CPMLDR_BIOS	
	.globl	_bios
; if it's a CPMLDR BIOS, do almost nothing ("CP/M 68 System Guide" page 10)
_init:
	move.l	#biosmsg,a0
	bsr	tx_str
	RTS
_bios:	; if called as CPMLDR BIOS - call this way
bios:
	ENDC
;	// AS_CPMLDR_BIOS


entry:	; here we get control when trap #3 executed
; if >=2 and <= 4 skip
	IFNE TRACE_ALL_BIOS_CALLS
	cmpi.b	#4,d0
	ble	entry_skip_trace
	move.l	a1,-(sp)
	move.l	d1,-(sp)
	move.l	d0,-(sp)
	move.l	#bioscall_msg,a0
	bsr	tx_str
	move.l	(sp)+,d0
	move.l	d0,-(sp)
	bsr	tx_w
	move.l	(sp)+,d0
	move.l	(sp)+,d1
	move.l	(sp)+,a1
entry_skip_trace:
	ENDC

	cmpi.w	#nfuncs,d0
	bcc	trapng
	andi.l	#$00FF,d0
	lsl	#2,d0	; multipling d0 x 4
	move.l	#biosbase,a0
	adda.l	d0,a0
	move.l	(a0),a0
	jsr	(a0)
trapng:
	IFNE	AS_CPMLDR_BIOS	
	rts	; if it's a CPMLDR BIOS, return with RTS
	ENDC
	IFEQ	AS_CPMLDR_BIOS	
	rte	; If not - return from a TRAP
	ENDC
; ******************* That was BIOS entry point, now the jump table ******************
biosbase:
	dc.l	_init
	dc.l	wboot
	dc.l	constat
	dc.l	conin
	dc.l	conout	; 4
	dc.l	lstout
	dc.l	pun
	dc.l	rdr
	dc.l	home
	dc.l	seldsk	; 9
	dc.l	settrk	; 10 A
	dc.l	setsec	; 11 B
	dc.l	setdma	; 12 C
	dc.l	read	; 13 D
	dc.l	write	; E
	dc.l	listst	; F
	dc.l	sectran	; 16 10
	dc.l	setdma	; 17 11
	dc.l	getseg	; getmrt,18 12
	dc.l	getiob	 ;19 13
	dc.l	setiob	; 20 14
	dc.l	flush	; 21 15
    dc.l	setexc	; 22 16

nfuncs	equ (*-biosbase)/4

;****************** BIOS Functions themselfs **************************************

wboot:
	IFNE	AS_CPMLDR_BIOS		; CPMLDR BEHAVES THIS WAY
	rts
	ENDC
	IFEQ	AS_CPMLDR_BIOS	
	jmp	_ccp
	ENDC

;*********************************************
constat:
;------------------ Test input for a char
; return 1 in d0 if true
	move.b      d0,(slowclock) ; switching clock slow
	move.b      (uartstat),d0
	move.b      d0,(fastclock) ; switching clock fast
	andi.b      #$02,d0 ; bit 1 is 1?
	lsr	#1,d0	; hope this will return 1 if is
	rts

;*********************************************
conin:
	bsr constat      ; waiting for char in input
	tst	d0
	beq	conin
	move.b      d0,(slowclock) ; switching clock slow
	move.b      (uartdata),d0   ; read from uart
	move.b      d0,(fastclock) ; switching clock fast
; in CP/M System guide
;	and.l	#$7F,d0
	rts

;*********************************************
conout:
	move.b      d0,(slowclock) ; switching clock slow
	move.b      (uartstat),d0	; getting UART TX READY status
	move.b      d0,(fastclock) ; switching clock fast
	andi.b	#$01,d0
	bne	conout			; not clear to send
	move.b      d0,(slowclock) ; switching clock slow
	move.b	d1,(uartdata)	; writing byte
	move.b      d0,(fastclock) ; switching clock fast
	rts
    

;*********************************************
lstout:
	rts
;*********************************************
pun:
	rts
;*********************************************
rdr:	
	rts

;*********************************************

home:
	clr.b	track
	rts

;*********************************************
seldsk: ; select disk given by register d1.b	
	moveq	#0,d0
	cmp.b	#maxdsk,d1
;*********************	bpl	selrtn
	bge	selrtn
	move.b	d1,seldrv
	move.b	d1,d0
	mulu	#dphlen,d0
	add.l	#dph0,d0	; returning correct dph
selrtn:	; return from seldsk
	rts

;*********************************************
; ACHTUNG ACHTUNG! IN SG, TRACK IS TWO BYTES!
; IN CP/M 68, looks like there is only 1 BYTE!
settrk:
	move.w	d1,track ; max 65535 (page 25 of system manual)
	rts
;*********************************************
setsec:
	move.w	d1,sector ; nothing said about maximum, but it is .w (page 26 of system manual)
	rts

;*********************************************
setdma:
	move.l	d1,dma
	rts

;*********************************************
; read one sector to dma address
; d0 == 00 if ok
read:
	IFNE AS_CPMLDR_BIOS
	move.b	#".",d1		; 
	bsr	conout			; 
	ENDC
	IFNE SD_SECTOR_DEBUG
	move.l	#reading_track_msg,a0
	bsr	tx_str
	move.b	seldrv,d1	; 
	bsr	tx_hex                  ;
	move.b	#":",d1		; 
	bsr	conout			; 
	move.w	track,d0	; 
	bsr	tx_w			; 
	move.b	#":",d1		; 
	bsr	conout			; 
	move.w	sector,d0	; 
	bsr	tx_w			; 
	move.l	#crlf_msg,a0	; 
	bsr	tx_str			; 
	ENDC

	bsr	setLBAaddr

	IFNE SD_READ_CACHE
; debug - delete it
;move.l	d0,-(sp)
;move.l	#reading_track_comparing,a0
;bsr	tx_str
;move.l	lba_in_sd_sector,d0 
;bsr	tx_l
;move.l	(sp)+,d0
; up to here

	cmp.l	lba_in_sd_sector,d0 ; was this sector
; 	read before?
	beq	readskip
	move.l	d0,-(sp) ; pushing lba to stack
;    move.l	#reading_track_msg_really,a0
;    bsr	tx_str
	ENDC

	bsr sd_read_sector_in_d0
	cmpi.w	#$0,d0
	beq	read_ok
; sector was not read ok
	IFNE SD_READ_CACHE
	move.l	(sp)+,d1	; popping lba from stacl
	move.l	#$FFFFFFFF,lba_in_sd_sector
	ENDC
; and by the way a0.w still holds 1
	rts


read_ok:

	IFNE SD_READ_CACHE
;    move.l	(sp)+,d0 ; lba_in_sd_sector ; poping lba from stack
;    move.l	d0,lba_in_sd_sector ;
;    bsr	tx_l
	move.l	(sp)+,lba_in_sd_sector ; poping lba from stack
; and storing it,as sector was read sucessfuly    
;	bra	readskipnomsg
readskip:
	IFNE SD_SECTOR_DEBUG
	move.l	#read_skip_msg,a0
	bsr	tx_str
	ENDC
;readskipnomsg:
	ENDC

	IFNE SD_SECTOR_DEBUG
	move.b	#'&',d1
	bsr	conout
	ENDC
	move.w	sector,d0
	andi.l	#3,d0

	IFNE SD_SECTOR_DEBUG
	move.l	d0,-(sp) ; showing two last bits
	bsr		tx_l
	move.l	(sp)+,d0
	ENDC

	lsl.l	#7,d0	; multiply 128

	IFNE SD_SECTOR_DEBUG
	move.l	d0,-(sp) ; showing offset
	bsr		tx_l
	move.l	(sp)+,d0
	ENDC

	add.l	#sd_sector,d0

	move.l	d0,a0	; our buffer + offset of number of cp/m sector
	IFNE SD_SECTOR_DEBUG
	move.b	#'%',d1
	bsr	conout
	ENDC


    IFNE SD_SECTOR_DEBUG
    move.l	a0,d0
    bsr	tx_l
    ENDC


    move.l	dma,a1  ; copying actually CP/M sector from our SD sector to 
; DMA
    move.w	#$7F,d4 ; length of CP/M sector size - 1
readtxb:
    move.b	(a0)+,(a1)+	; I mean (a0)->(a1)
    dbra	d4,readtxb

;    IFNE SD_SECTOR_DEBUG
;    move.l	#reading_track_returning_ok_msg,a0
;    bsr	tx_str
;    ENDC

;    clr.l	d0
    move.w	#0,d0
    rts


write: ; first, reading the whole track
    IFEQ	AS_CPMLDR_BIOS
    IFNE SD_SECTOR_DEBUG
;    IFNE SD_WRITE_DEBUG
    move.l	#reading_track_for_write_msg,a0
    bsr	tx_str
    move.b	seldrv,d1	; 
    bsr	tx_hex                  ;
    move.b	#":",d1		; 
    bsr	conout			; 
    move.w	track,d0	; 
    bsr	tx_w			; 
    move.b	#":",d1		; 
    bsr	conout			; 
    move.w	sector,d0	; 
    bsr	tx_w			; 
    move.l	#crlf_msg,a0	; 
    bsr	tx_str			; 
    ENDC

    bsr	setLBAaddr
    IFNE SD_READ_CACHE
    cmp.l	lba_in_sd_sector,d0 ; is the sector with this LBA allready in buffer?
    beq	writereadskip
    move.l	a0,-(sp) ; storing LBA to stack
    ENDC
; move.l	d0,d6	; we do not use d6, right? Nooo!


    bsr sd_read_sector_in_d0
    cmpi.w	#$0,d0
    beq	wread_ok
; sector was not read ok
    IFNE SD_READ_CACHE
    move.l	(sp)+,d1	; popping lba from stack
    move.l	#$FFFFFFFF,lba_in_sd_sector
    ENDC
    IFNE	SD_VERBOSE_ERRMSGS
    move.l	#w_cantread_sector,a0
    bsr	sd_verbose_msg
    ENDC
; and by the way a0.w still holds 1
    rts

wread_ok:
    IFNE SD_READ_CACHE
    move.l	(sp)+,lba_in_sd_sector ; poping lba from stack and storing it
; saying this sector was sucessfuly read
    bra	writereadskipnomsg
    ENDC


writereadskip:
    IFNE SD_SECTOR_DEBUG
    move.l	#read_skip_msg,a0
    bsr	tx_str
    ENDC

writereadskipnomsg:
;    IFNE SD_WRITE_DEBUG
    IFNE SD_SECTOR_DEBUG
    move.b	#'@',d1
    bsr	conout
    ENDC
    move.w	sector,d0
    andi.l	#3,d0

    IFNE SD_SECTOR_DEBUG
;    IFNE SD_WRITE_DEBUG
    move.l	d0,-(sp) ; showing two last bits
    bsr		tx_l
    move.l	(sp)+,d0
    ENDC

    lsl.l	#7,d0	; multiply 128

    IFNE SD_SECTOR_DEBUG
;    IFNE SD_WRITE_DEBUG
    move.l	d0,-(sp) ; showing offset
    bsr		tx_l
    move.l	(sp)+,d0
    ENDC

    add.l	#sd_sector,d0
    move.l	d0,a0	; our buffer + offset of number of cp/m sector

    IFNE SD_SECTOR_DEBUG
;    IFNE SD_WRITE_DEBUG
    move.b	#'#',d1
    bsr	conout
    move.l	a0,d0
    bsr	tx_l
    ENDC

; so, we have a0 poining to where we will copy dma
    move.l	dma,a1  
    move.w	#$7F,d4 ; length of CP/M sector size - 1
writtxb:
    move.b	(a1)+,(a0)+	; I mean (a1)->(a0)
    dbra	d4,writtxb

    IFNE SD_SECTOR_DEBUG
;    IFNE SD_WRITE_DEBUG
    move.l	#writing_track_msg,a0
    bsr	tx_str
    move.b	seldrv,d1	; 
    bsr	tx_hex                  ;
    move.b	#":",d1		; 
    bsr	conout			; 
    move.w	track,d0	; 
    bsr	tx_w			; 
    move.b	#":",d1		; 
    bsr	conout			; 
    move.w	sector,d0	; 
    bsr	tx_w			; 
    move.l	#crlf_msg,a0	; 
    bsr	tx_str			; 
    ENDC


; and now we have to write whole the sd_sector to sector
    bsr	setLBAaddr
    move.b	#SD_CMD24,d1
    bsr	sd_write_cmd_with_args	; sending command

; waiting for 00
;    move.w	#$7FFF,d5
;writloop3:
;    move.w	#$7FFF,d4
    move.w	#$40,d4
writloop31:
    IFNE SD_WRITE_DEBUG
    bsr	sd_read_byte_debug	; will return in D0
    ENDC
    IFEQ SD_WRITE_DEBUG
    bsr	sd_read_byte	; will return in D0
    ENDC

    cmpi.b	#$00,d0
    beq	writel3ex	; got ok to write
    dbra	d4,writloop31
;    dbra	d5,writloop3

    IFNE SD_READ_CACHE
    move.l	#$FFFFFFFF,lba_in_sd_sector
    ENDC
    
    IFNE	SD_VERBOSE_ERRMSGS
    move.l	#w_cantget0,a0
    bsr	sd_verbose_msg
    ENDC
    move.w	#1,d0	; cannot get 00 - "ready to write" mark
    rts

writel3ex: ; write loop 3 exit
    move.b	#$FF,d1
    bsr   sd_write_byte   ; one byte gap
    move.b	#$FE,d1
    bsr   sd_write_byte   ; starting data

    move.l	#sd_sector,a0	; writing SD sector itself
    move.w	#$1FF,d4	; #200 bytes of sector data -1
writloop4:
    move.l	d4,-(sp) ;DELETE THIS PUSH?
    move.b	(a0)+,d1
    IFNE SD_WRITE_DEBUG
    bsr sd_write_byte_quiet
    ENDC
    IFEQ SD_WRITE_DEBUG
    bsr	sd_write_byte
    ENDC
    move.l	(sp)+,d4 ; DELETE THIS POP?
    dbra	d4,writloop4

; here we transferred whole 1 SD track (4 CPM tracks) in sd_sector
    clr.b	d1
    bsr	sd_write_byte
    clr.b	d1
    bsr	sd_write_byte	; checksumm - 2 bytes, noone cares

; waiting SD to accept sector
;    move.w	#$7FFF,d5
;writeacc1:
;    move.w	#$7FFF,d4
    move.w	#$7F,d4
writeaccl2:
    move.l	d5,-(sp) ;DELETE THIS PUSH?
    move.l	d4,-(sp) ;DELETE THIS PUSH?
    IFNE SD_WRITE_DEBUG
    bsr	sd_read_byte_debug	; will return in D0
    ENDC
    IFEQ SD_WRITE_DEBUG
    bsr	sd_read_byte	; will return in D0
    ENDC
    move.l	(sp)+,d4 ; DELETE THIS POP?
    move.l	(sp)+,d5 ; DELETE THIS POP?
    andi.b	#$E,d0
    cmpi.b	#4,d0
    beq	writeaccept
    dbra	d4,writeaccl2
;    dbra	d5,writeacc1
    IFNE SD_READ_CACHE
    move.l	#$FFFFFFFF,lba_in_sd_sector ; Wrong write
    ENDC
    IFNE	SD_VERBOSE_ERRMSGS
    move.l	#w_blocknotaccepted,a0
    bsr	sd_verbose_msg
    ENDC
    move.w	#1,d0	; data was not accepted
    rts



writeaccept:
;    move.w	#$7FFF,d5
;writefinl:
;    move.w	#$7FFF,d4 ; and now just wait for 0
    move.w	#$80,d4 ; and now just wait for 0
writefinl2:
    move.l	d5,-(sp) ;DELETE THIS PUSH?
    move.l	d4,-(sp) ;DELETE THIS PUSH?
;    bsr	sd_read_byte
    IFNE SD_WRITE_DEBUG
    bsr	sd_read_byte_debug	; will return in D0
    ENDC
    IFEQ SD_WRITE_DEBUG
    bsr	sd_read_byte	; will return in D0
    ENDC
    move.l	(sp)+,d4 ; DELETE THIS POP?
    move.l	(sp)+,d5 ; DELETE THIS POP?
    cmpi.b	#0,d0
    beq	write_ok
    dbra	d4,writefinl2
;    dbra	d5,writefinl

    IFNE SD_READ_CACHE
    move.l	#$FFFFFFFF,lba_in_sd_sector
    ENDC
    IFNE	SD_VERBOSE_ERRMSGS
    move.l	#w_didnotwrite,a0
    bsr	sd_verbose_msg
    ENDC
    move.w	#1,d0	;could not wait for data accpt
    rts

; while we have 0 - skip it (and make some huge timeout here; and catch errors!)
write_ok:
    IFNE SD_WRITE_DEBUG
    bsr	sd_read_byte_debug	; will return in D0
    ENDC
    IFEQ SD_WRITE_DEBUG
    bsr	sd_read_byte	; will return in D0
    ENDC
    move.b	d0,d0
    beq	write_ok



;CMD13 *********************************************
    IFNE SD_CMD13_AFTER_WRITE

    move.b	#SD_CMD13,d1	; query status of prev command 
    clr.l	d0
    bsr	sd_write_cmd_with_args	; sending command
    move.w	$80,d4
write_ok_l:
;    IFNE SD_WRITE_DEBUG

    IFNE SD_WRITE_DEBUG
    bsr	sd_read_byte_debug	; will return in D0
    ENDC
    IFEQ SD_WRITE_DEBUG
    bsr	sd_read_byte	; will return in D0
    ENDC
    cmpi.b	#$ff,d0
    bne	write_ok_l_e
    dbra	d4,write_ok_l
    IFNE SD_READ_CACHE
write_ok_l_no:
    move.l	#$FFFFFFFF,lba_in_sd_sector	; cannot read CMD13 reply
    ENDC
    IFNE	SD_VERBOSE_ERRMSGS
    move.l	#w_cmd13no,a0
    bsr	sd_verbose_msg    
    ENDC

    move.w	#1,d0	;could not wait for data accpt
    rts


write_ok_l_e:
    cmpi.b	#$00,d0
    bne	write_ok_l_no   ; that should be 0
    IFNE SD_WRITE_DEBUG
    bsr	sd_read_byte_debug	; will return in D0
    ENDC
    IFEQ SD_WRITE_DEBUG
    bsr	sd_read_byte	; will return in D0
    ENDC
    cmpi.b	#$0,d0	; Read CMD13 "R2" status, but it is not 0 0
    beq	write_ok_fin
    IFNE	SD_READ_CACHE
    move.l	#$FFFFFFFF,lba_in_sd_sector	; cannot read CMD13 reply
    ENDC
    IFNE	SD_VERBOSE_ERRMSGS
    move.l	#w_cmd13status,a0
    bsr	sd_verbose_msg    
    ENDC
    move.w	#1,d0	;could not wait for data accpt
    rts



write_ok_fin:
    IFNE SD_READ_CACHE ;
    bsr	setLBAaddr	; the write is ok, recalculating LBA
    move.l	d0,lba_in_sd_sector ; and storing it
    ENDC
;    vv CMD13_AFTER_WRITE
    ENDC
;///CMD13 *********************************************



;    bsr	sd_read_byte_debug	; will return in D0
;    bsr	sd_read_byte_debug	; will return in D0
;    bsr	sd_read_byte_debug	; will return in D0
;    bsr	sd_read_byte_debug	; will return in D0

    clr.l	d0
    ENDC
;***^^^ This is if not CPMLDR
    rts

;*********************************************
listst:
    move.b	#$ff,d0
    rts
;*********************************************
sectran:		; return in D0
    move.l	d1,d0	; sector D1 translated via table pointed by d2
    rts
;setdma:
getseg:
    move.l	#memrgn,d0
    rts
getiob:
setiob:
    rts

flush:
    clr.l	d0
    rts

setexc:	; installing interrupt vector
;   move.l	d1,-(sp)
;    move.b	#$0D,d1
;    bsr	conout
;    move.b	#$0A,d1
;    bsr	conout
;    move.b	#'&',d1
;    bsr	conout
;    move.b	#'&',d1
;    bsr	conout
;    move.l	(sp)+,d1
    andi.l	#$FF,d1	; 
    lsl	#2,d1	; multiply *4
    movea.l	d1,a0	
    move.l	(a0),d0
    move.l	d2,(a0)
;    move.b	#$0D,d1
;    bsr	conout
;    move.b	#$0A,d1
;    bsr	conout
;    move.b	#'*',d1
;    bsr	conout
;    move.b	#'*',d1
;    bsr	conout
    rts
; My function of calculating LBA of track/sector/seldrv
setLBAaddr:
    clr.l	d4	; Here will be LBA
    clr.l	d0
    move.w      sector,d0 ; we have $80 CP/M sec/track
    andi.w	#$FC,d0	; 4 CP/M track for our SD 512 sec/track
    lsl.l	#7,d0
    add	d0,d4	; .w? looks ok

    clr.l	d0	; adding track
    move.w	track,d0
    swap	d0
    lsr.l	#2,d0
    add.l 	d0,d4


    clr.l	d0
    move.b	seldrv,d0
    swap	d0
    lsl.l	#7,d0

    add.l	d0,d4	; have everything in d4; why not addl.l d4.d0?

    IFNE SD_SECTOR_DEBUG
    move.l	d4,d0
    bsr	tx_l
    ENDC

    move.l	d4,d0
        
    rts

; waits for byte in d0 d1 times with error message in a0
wait_for_d0:
    IFNE SD_WRITE_DEBUG
    bsr	sd_read_byte_debug	; will return in D0
    ENDC
    IFEQ SD_WRITE_DEBUG
    bsr	wait_for_d0_ok	; will return in D0
    ENDC
    dbra	d4,wait_for_d0
    IFNE SD_READ_CACHE
    move.l	#$FFFFFFFF,lba_in_sd_sector
    ENDC
    IFNE	SD_VERBOSE_ERRMSGS
    bsr	sd_verbose_msg
    ENDC
    move.w	#1,d0	;could not wait for data accpt
    rts
wait_for_d0_ok:
;    clr.w	d0
    rts

; writes SD command D1, with  arguments D0
sd_write_cmd_with_args:

; should be there "write FF"? That s be good
  move.l	d0,d7
  move.l	d1,d6
  move.b	#$FF,d1		; lets do it several times
  bsr   sd_write_byte   ; FF
;  move.b	#$FF,d1
;  bsr   sd_write_byte   ; FF
;  move.b	#$FF,d1
;  bsr   sd_write_byte   ; FF
;  move.b	#$FF,d1
;  bsr   sd_write_byte   ; FF
;  move.b	#$FF,d1
;  bsr   sd_write_byte   ; FF
;  move.b	#$FF,d1
;  bsr   sd_write_byte   ; FF
;  move.b	#$FF,d1
;  bsr   sd_write_byte   ; FF
;  move.b	#$FF,d1
;  bsr   sd_write_byte   ; FF

  move.l	d6,d1
  bsr   sd_write_byte   ; command

  move.l	d7,d1
  rol.l	#8,d1
  move.l	d1,d7
  andi.b	#$ff,d1
  bsr   sd_write_byte   ; first byte of arguments

  move.l	d7,d1
  rol.l	#8,d1
  move.l	d1,d7
  andi.b	#$ff,d1
  bsr   sd_write_byte   ; second byte of arguments

  move.l	d7,d1
  rol.l	#8,d1
  move.l	d1,d7
  andi.b	#$ff,d1
  bsr   sd_write_byte   ; third byte of arguments

  move.l	d7,d1
  rol.l	#8,d1
  move.l	d1,d7
  andi.b	#$ff,d1
  bsr   sd_write_byte   ; forth byte of arguments

  move.b	#$95,d1
  bsr   sd_write_byte   ; second byte of arguments
  rts

; writes byte quiet 
  IFNE SD_WRITE_DEBUG
sd_write_byte_quiet:
  move.l        d3,-(sp)
  bra sd_write_byte_skip_ascii
  ENDC

; writes byte D1
sd_write_byte:
;  andi.l       #$000000FF,d1 ; that should be byte ; but looks like thats not needed
  move.l        d3,-(sp)

; delete this debug code
  IFNE SD_WRITE_DEBUG
  move.b	d1,-(sp)
  move.b                #'[',d1
  bsr   conout
  move.b                #'W',d1
  bsr   conout
  move.b                #':',d1
  bsr   conout
  move.b	(sp)+,d1
  move.b	d1,-(sp)
  bsr   tx_hex
  move.b	(sp)+,d1
  cmpi.b	#$20,d1
  ble	sd_write_byte_skip_ascii
  bsr	conout
sd_write_byte_skip_ascii:
  ENDC
; down to here
  move.w                #7,d3   ; 8 bits in byte

sd_write_bit_loop:

    move.b                #0,d0
    add.b         d1,d1
    addx.b                d0,d0   ; should add X==Carry flag,
    add.b         d0,d0   ; 2 or 0
    move.b                d0,sd_port ; cs=0 data=data clock=0
    add.b         #$1,d0
    move.b                d0,sd_port ; cs=0 data=data clock=1
    subi.b        #$1,d0
    move.b                d0,sd_port ; cs=0 data=data clock=0
    dbra  d3,sd_write_bit_loop


  IFNE SD_WRITE_DEBUG
  move.b                #']',d1
  bsr   conout
  ENDC

  move.l        (sp)+,a3
  rts

    

; returns on D0
sd_read_byte:  
    move.l	d3,-(sp)
    move.w                #7,d3

    move.b                #$0,d0  ; here will be byte stored
sd_read_byte_read_bit:
    add.b         d0,d0   ; shifting it left
    move.b                #3,sd_port      ; cs=0,data=1,clk=1

    move.b                sd_port,d1      ; reading inverted data
    andi.b                #1,d1   ; the lower bit is input
    eori.b                #1,d1   ; we have an inverted data

    add.b         d1,d0   ; adding to result
    move.b                #2,sd_port      ; cs=0,data=1,clk=0

    
    dbra  d3,sd_read_byte_read_bit
    move.l	(sp)+,d3
    rts

  IFNE SD_WRITE_DEBUG
sd_read_byte_debug:
    bsr	sd_read_byte
    move.b                d0,-(sp)
    move.b                #'(',d1
    bsr   conout
    move.b                #'R',d1
    bsr   conout
    move.b                #':',d1
    bsr   conout
    move.b                (sp)+,d1
    move.b                d1,-(sp)
    bsr   tx_hex
    move.b                #')',d1
    bsr   conout
    move.b                (sp)+,d0
;  move.l	(sp)+,d3
    rts
    ENDC



;
; Reads SD sector. Input - LBA in A as setLBAaddr returns it.
; Output - d0.w = 1 - is a error, d0.w = 0 - no error
sd_read_sector_in_d0:

    move.b	#SD_CMD17,d1
    bsr	sd_write_cmd_with_args

;    move.w	#$7F,d5	; waiting for command accepted
;sd_rs_readloop01:
    move.w	#$7F,d4	; waiting for command accepted
sd_rs_readloop0:
    IFNE SD_WRITE_DEBUG
    bsr	sd_read_byte_debug	; will return in D0
    ENDC
    IFEQ SD_WRITE_DEBUG
    bsr	sd_read_byte	; will return in D0
    ENDC
    cmpi.b	#0,d0
    beq	sd_rs_readloop0e
    dbra	d4,sd_rs_readloop0
;    move.b	#'.',d1
;    bsr	conout
;    dbra	d5,sd_rs_readloop01


    IFNE	SD_VERBOSE_ERRMSGS
    move.l	#r_cmd17no0,a0
    bsr	sd_verbose_msg
    ENDC
    move.w	#1,d0	; cannot get data token in 101 iterations, reporting error
    rts

sd_rs_readloop0e:
    IFNE SD_SECTOR_DEBUG
    move.b	d0,d1
    bsr	tx_hex
    ENDC

    move.w	#$80,d4
sd_rs_readloop1:
    move.w	#$7FFF,d5
sd_rs_readloop12:
    IFNE SD_WRITE_DEBUG
    bsr	sd_read_byte_debug	; will return in D0
    ENDC
    IFEQ SD_WRITE_DEBUG
    bsr	sd_read_byte	; will return in D0
    ENDC
    cmpi.b	#$FE,d0
    beq	sd_rs_readl1ex	; got data token
    dbra	d5,sd_rs_readloop12
    move.b	#'.',d1
    bsr	conout
    dbra	d4,sd_rs_readloop1
    IFNE	SD_VERBOSE_ERRMSGS
    move.l	#r_cmd17nodatatoken,a0
    bsr	sd_verbose_msg
    ENDC
    move.w	#1,d0	; cannot get data token in 101 iterations, reporting error
    rts

sd_rs_readl1ex:	; read loop 1 exit

    move.l	#sd_sector,a0
    move.w	#$201,d4	; #200 bytes of sector data, 2 bytes of CRC -1
sd_rs_readloop2:
    bsr	sd_read_byte
    move.b	d0,(a0)+
    dbra	d4,sd_rs_readloop2

;    bsr	sd_read_byte_debug
;    bsr	sd_read_byte_debug
;    bsr	sd_read_byte_debug

    clr.w	d0	; All good 
    rts


	IFNE	SD_VERBOSE_ERRMSGS
sd_verbose_msg:
	move.b	(a0)+,d0
	bsr	tx_str
	move.l	(sp)+,d0
	rts
	ENDC

;*******************************************************************
; My own local functions for debug
; tx string from A0
tx_str:	
	move.b	(a0)+,d1
	beq	tx_str_exit
	move.l	a0,-(sp)
	bsr	conout
	move.l	(sp)+,a0
	bra	tx_str
tx_str_exit:
	rts

;*********************************************

; Tx byte in D1 (spoils D0)
tx_hex:
    move.b      d0,(slowclock) ; switching clock slow
    move.b      (uartstat),d0	; getting UART TX READY status
    move.b      d0,(fastclock) ; switching clock fast
    andi.b	#$01,d0
    bne	tx_hex			; not clear to send
    move.b      d0,(slowclock) ; switching clock slow
    move.b	d1,(uartstat)	; writing byte TO STATUS REGISTER - thats a HEX output
    move.b      d0,(fastclock) ; switching clock fast
    rts
; tx_l in D0
tx_l:
    move.l      d0,-(sp)
    swap        d0
    andi.l      #$FFFF,d0
    bsr tx_w
    move.l      (sp)+,d0
    andi.l      #$FFFF,d0
    bsr tx_w
    rts
;-------------------tx word (hex) , d0
tx_w:
    move.l      d1,-(sp)
    move.l      d0,-(sp)
; lsr should be here!
    ror #8,d0
    move.b	d0,d1
    bsr tx_hex
    move.l      (sp)+,d1
    andi.l      #$00FF,d1
    bsr tx_hex
    move.l      (sp)+,d1
    rts

	IFEQ AS_SRECORD
;	data
	ENDC

	IFNE SD_SECTOR_DEBUG
reading_track_msg:
	dc.b	$0D,$0A,'Reading track: ',0
reading_track_for_write_msg:
	dc.b	$0D,$0A,'Reading track for write: ',0
writing_track_msg:
	dc.b	$0D,$0A,'Writing track: ',0
reading_track_returning_ok_msg:
	dc.b	$0D,$0A,'Track Read OK!',$0D,$0A,0
read_skip_msg:
	dc.b	$0D,$0A,'(skipping read sector)',$0D,$0A,0
	ENDC

	IFNE TRACE_ALL_BIOS_CALLS
bioscall_msg:
	dc.b	$0D,$0A,'BIOS CALL:',0
	ENDC


	IFNE	SD_VERBOSE_ERRMSGS
w_cantread_sector:
	dc.b	$0D,$0A,'W:Cant read sector',0
w_cantget0:
	dc.b	$0D,$0A,'W:Cant get 0',0
w_blocknotaccepted:
	dc.b	$0D,$0A,'W:Write was not accepted',0
w_didnotwrite:
	dc.b	$0D,$0A,'W:Did not write',0
	IFNE	SD_CMD13_AFTER_WRITE
w_cmd13no:
	dc.b	$0D,$0A,'W:Cannot cmd13',0
w_cmd13status:
	dc.b	$0D,$0A,'W:Cmd13 status wrong',0
	ENDC
r_cmd17no0:
	dc.b	$0D,$0A,'R:Cmd17 cannot get 0',0
r_cmd17nodatatoken:
	dc.b	$0D,$0A,'R:Cmd17 cannot get data token',0
	ENDC

biosmsg:
	IFNE	AS_CPMLDR_BIOS
	dc.b	$0D,$0A,'DtZ 68kavr Loader BIOS 0.00.00',$0D,$0A,0
	ENDC
	IFEQ	AS_CPMLDR_BIOS
	IFNE AS_SRECORD
	dc.b	$0D,$0A,'DtZ 68kavr SRECORD BIOS 0.00.00',$0D,$0A,0
	ENDC
	IFEQ AS_SRECORD
	dc.b	$0D,$0A,'DtZ 68kavr DISK BIOS 0.00.00',$0D,$0A,0
	ENDC
	ENDC

crlf_msg:
	dc.b	$0D,$0A,0

seldrv	dc.b	0	; selected disk
	even
track	dc.w	0
sector	dc.w	0
dma	dc.l	0
	even
memrgn	dc.w	1
	dc.l	$400
	IFNE AS_SRECORD
	dc.l	$14c00 ; goes until 15000 in our case?
	ENDC
	IFEQ AS_SRECORD
	dc.l	$17c00 ; goes until 18000 hex 
	ENDC


	even

	IFNE SD_READ_CACHE
;reading_track_msg_really:
;	dc.b	'[RRREADING]',0
;reading_track_comparing:
;	dc.b	$0D,$0A,'Will compare to ',0

;cached_r_msg:
;	dc.b	'Cached block',0
	even
lba_in_sd_sector:
	dc.l	$FFFFFFFF
	ENDC

	include	"dph.asm"


sd_sector:
; if we assemble as SRecord, we do not need this 514 bytes in Srecord
	IFEQ AS_SRECORD
	ds.b	$202
	ENDC
sd_secte:
; here should be at least $202 bytes
	END
