//#define __AVR_LIBC_DEPRECATED_ENABLE__ 1
//#define F_CPU 12000000L
#define F_CPU 8000000L
#include <avr/io.h>
#include <avr/wdt.h>
#include <inttypes.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

//#define BAUD 9600
#define BAUD 38400
#define MYUBRR (F_CPU/(16L*BAUD))-1

#define DTZ_MAIN
#include "main.h"

// Function deinitions
void ATprocess();
void ATsayError(); 

#ifdef WITH_CTS
static uint8_t input_buffer[RX_BUF_SIZE];
static volatile uint8_t rcvd_char_ptr; // Pointer to place next recived from serial char
static volatile uint8_t read_char_ptr; // Pointer to next char to tell the CPU when it reads serial
#endif

static int input_buffer_p;
static uint8_t byte_in_rx_que; /* byte in input que. Must be 0x02 if is */
static uint8_t byte_came; /* actual byte */

/* Bootmode tells INT0 handle what to do
on bootmode=1 - work as program feeder
on bootmode=0 - work as bus spy
*/
volatile char bootmode;
volatile char writing_on_bus = 0;
/* static */
volatile long int reset_ticks = 0;

/* function for long delay */
void delay_ms(uint16_t ms) {
        while ( ms )
        {
                _delay_ms(1);
                ms--;
        }
}


#ifdef WITH_CTS
int RX_buffer_is_full()
{ // and unable to get free chars ; returns 1 if is
 return(rcvd_char_ptr == ((read_char_ptr - 1) % RX_BUF_SIZE ));
}
void RX_push_to_buffer(uint8_t c)
{ // All clear with it, right?
 rcvd_char_ptr = ( (rcvd_char_ptr+1) & RX_BUF_MASK /* % RX_BUF_SIZE */ );
 input_buffer[rcvd_char_ptr] = c;
}
uint8_t RX_pop_from_buffer()
{ // see comment on RX_push_to_buffer
 read_char_ptr = (read_char_ptr+1) & RX_BUF_MASK /*% RX_BUF_SIZE */ ;  
 return(input_buffer[read_char_ptr]);
}
int RX_free_bytes_in_buffer()
{
 return(( RX_BUF_SIZE - 1 - rcvd_char_ptr + read_char_ptr) & RX_BUF_MASK /* % RX_BUF_SIZE */);
}
#endif

ISR(USART_RXC_vect) 
{
 static uint8_t c; 

 /* if we get a BREAK from serial line - reset! */
 if (UCSRA & (1 << FE ))
 {
  static unsigned const PROGMEM char b[] = "\n\r**GOT BREAK**\n\r";
  USART_PUTS_my(b);

  wdt_enable(WDTO_30MS);
  while(1) {};
 }

#ifdef WITH_CTS
// c = (uint8_t)c1;
 c = UDR; // Getting a char from serial
 if (!RX_buffer_is_full())
 {
  RX_push_to_buffer(c);
  if ( RX_free_bytes_in_buffer()  <  (RX_BUF_SIZE / 4) ) // if we have less then 1/4 of free bytes in buffer
 // if ( RX_free_bytes_in_buffer()  <  (RX_BUF_SIZE / 2) ) // if we have less then 1/4 of free bytes in buffer
  {
   rts_down(); // saying we do not want them more for now
  }
 }
#else
 c = UDR;
 byte_in_rx_que = 0x02;
 byte_came = c;
#endif
}

void ATsayError() 
{
 static unsigned const PROGMEM char b[] = "\n\rERROR\n\r";
 USART_PUTS_my(b);
}

void ATprocess()
{
  static unsigned const PROGMEM char b[] = "\n\rVOBLA!\n\rOK\n\r";
  reset_ticks = 100;
  USART_PUTS_my(b);
}


/***************************** External interrupt handling *****************/
/* this interrupt is called when MCU reads or writes anywhere  in bootmode*/
ISR (INT0_vect)
{
#ifdef DEBUG
 static unsigned  const PROGMEM char b1[] = "\n\rINT: ";
#endif
 static unsigned char rdwr = 0;
 int b;
 rdwr = (RDWR_PORTIN & _BV(RDWR_PIN));
#ifdef DEBUG
// USART_PUTS_my(b1);
#endif
 if (rdwr) { /* reads */
  if (bootmode) 
  { /* bytefeeder, and notify main loop to stop writing to data bus on /DS goes up */
   if (!reset_ticks) {
#ifdef DEBUG
    USART_PUTCH('R');
#endif
    b = bytefeeder();
#ifdef DEBUG
    uart_puthex_byte((unsigned char)b);
#endif
    if (b != -1) // Normal byte
    {
    //writing_on_bus = 1; (is set by writebyte)
     writebyte((unsigned char)b);
    } else { // Boot mode ended!
//    bootmode = 0;
     exit_bootmode();
     bootmode = 0;
     data_bus_to_read();
     GICR &= ~(1 << INT0);      // Turns off INT0
     reset_CPU();
    }
   } // If we got a fake read during reset
  } else 
  {  // spy mode
#ifdef DEBUG
   USART_PUTCH('r');
   uart_puthex_byte(read_data_bus());
#endif
  }
 } else {
#ifdef DEBUG
  USART_PUTCH('W');
#endif
  data_bus_to_read(); // It should allready be there
#ifdef DEBUG
  uart_puthex_byte(read_data_bus());
#endif
 }
}

/* this happens when CPU read/writes to serial */
ISR (INT1_vect)
{
#ifdef DEBUG
 static unsigned const  PROGMEM char b1[] = "\n\rINT1";
 static unsigned const PROGMEM char cpu_speaks[] = "\n\rCPU SPEAKS:";
 static unsigned const PROGMEM char cpu_reads[] = "\n\rCPU READS:";
// static unsigned PROGMEM char cpu_ready[] = "\n\rREADY:";
 static unsigned const PROGMEM char cpu_notready[] = "\n\rNOT READY:";
 static unsigned const PROGMEM char charin_msg[] = "\n\rBYTE FROM LINE:";
 static unsigned const PROGMEM char status_msg[] = "\n\rSENDING STATUS:";
#endif

 static unsigned char rdwr = 0;
 static unsigned char reads_status = 0;
 static unsigned char ready_to_write = 0;
 rdwr = (RDWR_PORTIN & _BV(RDWR_PIN));
 if (!rdwr) 
 { /* writes */
#ifdef DEBUG
  USART_PUTS_my(cpu_speaks);
  uart_puthex_byte(read_data_bus());
#endif
  if ( reads_status = (A1_PORTIN & _BV(A1_PIN)) ) {
// if 1, output HEX
   uart_puthex_byte(read_data_bus());
  } else {
//  uart_puthex_byte(read_data_bus());
   USART_PUTCH(read_data_bus());
  }
 } else
 { /* reads */
#ifdef DEBUG
  USART_PUTS_my(cpu_reads);
#endif
  if ( reads_status = (A1_PORTIN & _BV(A1_PIN)) )
  {
  /* if 1, reads status register; if not, read next char */
  // USART_READY_TO_SEND()) Returns 0 if ready
#ifdef DEBUG
   USART_PUTS_my(status_msg);
#endif

//   uart_puthex_byte( USART_READY_TO_SEND() | byte_in_rx_que);

#ifndef WITH_CTS
   writebyte( USART_READY_TO_SEND() | byte_in_rx_que); /* byte_ready is 0x02 if there is a byte in rx que */
#else
   writebyte( USART_READY_TO_SEND() | (read_char_ptr != rcvd_char_ptr ? 0x02 : 0) ); 
//   if (read_char_ptr != rcvd_char_ptr) USART_PUTCH('#'); else USART_PUTCH('%');
     /* byte_ready is 0x02 if there is a byte in rx que */
#endif

  } else 
  {
#ifdef DEBUG
   USART_PUTS_my(charin_msg);
#endif
//   uart_puthex_byte(byte_came);

#ifndef WITH_CTS
   writebyte(byte_came);
   byte_in_rx_que = 0; /* say rx que is empty */
#else
  writebyte(RX_pop_from_buffer());

  if ( RX_free_bytes_in_buffer() > /* 3*(RX_BUF_SIZE/4) */ RX_BUF_SIZE - 2 ) // If RX Buffer is almost empty
  {
   rts_up();  // we have some space for chars
  }
#endif
  }
 }
 /* Here be serial IO */
}
/*
#ifdef WITH_WATCHDOG
ISR(TIMER1_COMPA_vect)
{
   USART_PUTCH('%');
}
#endif
*/
/***************************** Main program ********************************/
int main()
{
 /************ Set Data bus to High impedance, init clock, reset lines ***********/
 setioports();
 /*********************************** Initing timer ******************************/

/*
 TIMSK = _BV(OCIE1A); // Enabling interrupt

 TCCR1B = _BV(CS12)    // 256 prescale 
         | _BV(WGM12);  // CTC mode, TOP = OCR1A
// OCR1A = 825;        // count up to TOP   1hz with 8 meg system clock BLINK FAST
// OCR1A = 15825;        // count up to TOP   1hz with 8 meg system clock BLINK FAST
 OCR1A = 5825;        // count up to TOP   1hz with 8 meg system clock BLINK FAST
*/

 /*********************************** Com port settings ******************************/
 USART_Init(MYUBRR); 

 UCSRB |= (1 << RXCIE); // enable UDRempty intr. 

 /**************************** Installing interrupt handles ******************************/
/* https://sites.google.com/site/qeewiki/books/avr-guide/external-interrupts */
 /* Setting IRQ0 handling - write bytes in bootmode */
 DDRD &= ~(1 << DDD2);      // PD2 (INT0 pin) is now an input
 PORTD |= (1 << PORTD2);    // turn On the Pull-up

 MCUCR &= ~(1 << ISC00);   // Oh no, falling edge only
 MCUCR |= (1 << ISC01); 

 GICR |= (1 << INT0);      // Turns on INT0
 /* Done setting IRQ0 */

 /* Setting IRQ1 handling - DATA IO  */
 DDRD &= ~(1 << DDD3);      // PD3 (INTI pin) is now an input
 PORTD |= (1 << PORTD3);    // turn On the Pull-up
 MCUCR &= ~(1 << ISC10);   // Installing INT0 on falling edge
 MCUCR |= (1 << ISC11);
 GICR |= (1 << INT1);
/* Done setting IRQ1 */
#ifdef WITH_CTS
 rcvd_char_ptr = read_char_ptr = 0;
/* we have no chars in input que */
#endif
/*
#ifdef WITH_WATCHDOG
 TIMSK = _BV(OCIE1A); // Timer stuff
// TCCR1B = _BV(CS12)    // 256 prescale
 TCCR1B = _BV(CS12)    // 256 prescale
	 | _BV(CS10)
         | _BV(WGM12);  // CTC mode, TOP = OCR1A
 OCR1A = F_CPU/(2*1024);        // count up to TOP   1hz with 8 meg system clock BLINK FAST
#endif
*/
 enter_bootmode();
 input_buffer_p = 0;
 sei(); // Starting interrupts and running main program
 reset_CPU();

 while (1)
 {
#ifdef DEBUG
  _delay_us(1000);
#endif
  if (reset_ticks != -1) /* if not halted forever */
   click_clock(); /* click the clocks */

  // If we are [still] writing to bus, but DS 
  // (the same line that causes INT0) line gone
  // high - put data bus to read
  if (writing_on_bus && (
         (PIND & _BV(PIN2)) != 0)) {
   data_bus_to_read(); // Also sets writing_on_bus to 0
//#ifdef DEBUG
//   USART_PUTS_my(releasingbusmsg);
//#endif
  }


  if (reset_ticks)  // We are [still] doing reset
  {
   if (reset_ticks != -1) /* if not reset forever */
   {
    reset_ticks--;
    if ( reset_ticks == 0 ) { // stop doing it
     release_reset_CPU();
    }
   }
  }

 }
 return((int)0);
}
